﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.IO;
using System.Threading;
using Phidgets;

namespace StepperMidi
{
    public partial class Form1 : Form
    {
        Manager man;
        List<Stepper> steppers = new List<Stepper>();
        public Form1()
        {
            InitializeComponent();
        }

        private void loadToolStripMenuItem_Click(object sender, EventArgs e)
        {
            try
            {
                DialogResult res = openFileDialog1.ShowDialog();
                if (res == DialogResult.OK)
                {
                    if (openFileDialog1.FileName != null)
                    {
                        FileInfo file = new FileInfo(openFileDialog1.FileName);
                        if (file.Exists)
                        {
                            loadFile(file);
                            parseMidi();
                        }
                        else
                            throw new Exception("File doesn't exist");
                    }
                    else
                        throw new Exception("No file chosen");
                }
            }
            catch (Exception ex)
            {
                MessageBox.Show("Error: " + ex.Message, "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
                midiData = null;
                midi = null;
            }
        }

        private void exitToolStripMenuItem_Click(object sender, EventArgs e)
        {
            this.Close();
        }

        MemoryStream midiData = null;
        private void loadFile(FileInfo file)
        {
            using (FileStream fileStream = File.OpenRead(file.FullName))
            {
                midiData = new MemoryStream();
                midiData.SetLength(fileStream.Length);
                fileStream.Read(midiData.GetBuffer(), 0, (int)fileStream.Length);
            }
        }

        MidiFile midi = null;
        private void parseMidi()
        {
            if (midiData != null)
            {
                midiData.Position = 0;
                byte[] header = new byte[14];
                midiData.Read(header, 0, 14);

                if (Encoding.ASCII.GetString(header, 0, 4) != "MThd")
                {
                    throw new Exception("Bad format midi file");
                }

                int headerLength = (header[4] << 24) + (header[5] << 16) + (header[6] << 8) + header[7];
                if (headerLength != 6)
                {
                    throw new Exception("Unexpecte header length");
                }

                MIDIFormat format = (MIDIFormat)Enum.Parse(typeof(MIDIFormat), ((header[8] << 8) + header[9]).ToString());

                int numTracks = (header[10] << 8) + header[11];
                int division = (Int16)(header[12] << 8) + header[13];

                if (division < 0)
                    throw new Exception("I don't understand negative tempo");

                midi = new MidiFile();
                midi.speed = division;
                midi.tracks = new MidiTrack[numTracks];
                for (int i = 0; i < numTracks; i++)
                {
                    byte[] trackHeader = new byte[8];
                    midiData.Read(trackHeader, 0, 8);

                    if (Encoding.ASCII.GetString(trackHeader, 0, 4) != "MTrk")
                    {
                        throw new Exception("Bad format midi track in file");
                    }

                    int trackLength = (trackHeader[4] << 24) + (trackHeader[5] << 16) + (trackHeader[6] << 8) + trackHeader[7];

                    byte[] trackData = new byte[trackLength];
                    midiData.Read(trackData, 0, trackLength);

                    MidiTrack track = new MidiTrack(trackData, midi.tempo);
                    midi.tracks[i] = track;

                    if (track.tempo != 0)
                        midi.tempo = track.tempo;
                }
            }

            int i1=0;
            trackCmb.Items.Clear();
            foreach (MidiTrack track in midi.tracks)
            {
                if (track.noteData.Count > 0)
                {
                    trackCmb.Items.Add(i1.ToString());
                }
                i1++;
            }
            trackCmb.SelectedIndex = 0;
        }

        List<Thread> playerThreads = new List<Thread>();
        private void startButton_Click(object sender, EventArgs e)
        {
            stopPlayerThreads();
            if (midi != null)
            {
                int i2 = 0;
                List<PlayerObject> playerObjects = new List<PlayerObject>();

                int channelStart = midi.tracks[trackIndex].channels.IndexOf(channelIndex);
                for (int j = trackIndex; j < midi.tracks.Length; j++)
                {
                    if (midi.tracks[j].channels.Count == 0)
                        continue;
                    for (int k = channelStart; k < midi.tracks[trackIndex].channels.Count; k++)
                    {
                        if (i2 >= steppers.Count)
                            goto done;

                        //Map channels to steppers
                        PlayerObject play = new PlayerObject();
                        play.channel = midi.tracks[j].channels[k];
                        play.track = midi.tracks[j];
                        play.stepper = steppers[i2];
                        play.midi = midi;

                        playerObjects.Add(play);
                        playerThreads.Add(new Thread(new ParameterizedThreadStart(playMidiThreadFuntion)));
                        i2++;
                    }
                    channelStart = 0;
                }
            done:

                for(int i=0;i<playerThreads.Count;i++)
                    playerThreads[i].Start(playerObjects[i]);
            }
        }

        private void stopButton_Click(object sender, EventArgs e)
        {
            stopPlayerThreads();
        }

        void stopPlayerThreads()
        {
            foreach(Stepper stepper in steppers)
                stepper.steppers[0].Engaged = false;
            stopFlag = true;
            foreach (Thread playerThread in playerThreads)
            {
                if (playerThread != null && playerThread.IsAlive)
                {
                    playerThread.Join();
                }
            }
            foreach(Stepper stepper in steppers)
                stepper.steppers[0].Engaged = false;

            stopFlag = false;
            playerThreads.Clear();
        }

        Boolean stopFlag = false;
        int trackIndex = 0;
        int channelIndex = 0;
        void playMidiThreadFuntion(Object obj)
        {
            PlayerObject play = (PlayerObject)obj;

            //Play it
            foreach (NoteEvent note in play.track.noteData)
            {
                if (stopFlag)
                    return;

                int sleepTime = (int)(note.duration / play.midi.speed);

                if (note.channel == play.channel)
                {
                    if (note.eventType == MidiEvent.NoteOn)
                    {
                        //turn off briefly so we have space between notes
                        //stepper.steppers[0].Engaged = false;
                        //Thread.Sleep(50);
                        //sleepTime -= 50;

                        //Engage stepper
                        double frequency = Math.Pow(2, ((note.data[0] - 69) / 12.0)) * 440;
                        //Console.WriteLine("On: " + frequency.ToString());
                        try
                        {
                            //play.stepper.steppers[0].VelocityLimit = frequency;
                            //play.stepper.steppers[0].VelocityLimit = frequency * 8;
                            play.stepper.steppers[0].VelocityLimit = frequency * 16;
                            //play.stepper.steppers[0].VelocityLimit = frequency * 32;
                            //play.stepper.steppers[0].VelocityLimit = frequency * 64;
                            //play.stepper.steppers[0].VelocityLimit = frequency * 128;
                            play.stepper.steppers[0].Engaged = true;
                        }
                        catch { }
                    }
                    else if (note.eventType == MidiEvent.NoteOff)
                    {
                        //Disengage stepper
                        //Console.WriteLine("Off");
                        if(noOff == false)
                            play.stepper.steppers[0].Engaged = false;
                    }
                }

                while (sleepTime > 0)
                {
                    if(sleepTime < 250)
                    {
                        Thread.Sleep(sleepTime);
                        sleepTime=0;
                    }
                    else
                    {
                        Thread.Sleep(250);
                        sleepTime-=250;
                    }
                    if (stopFlag)
                        return;
                }
            }
        }

        private void trackCmb_SelectedIndexChanged(object sender, EventArgs e)
        {
            trackIndex = int.Parse(trackCmb.SelectedItem.ToString());
            channelCmb.Items.Clear();
            foreach (int channel in midi.tracks[trackIndex].channels)
                channelCmb.Items.Add(channel);
            channelCmb.SelectedIndex = 0;
        }

        private void Form1_Load(object sender, EventArgs e)
        {
            man = new Manager();
            man.Attach += new Phidgets.Events.AttachEventHandler(man_Attach);
            man.open();
        }

        void man_Attach(object sender, Phidgets.Events.AttachEventArgs e)
        {
            if (e.Device.Class == Phidget.PhidgetClass.STEPPER)
            {
                Stepper stepper = new Stepper();
                stepper.Attach += new Phidgets.Events.AttachEventHandler(stepper_Attach);
                stepper.Detach += new Phidgets.Events.DetachEventHandler(stepper_Detach);
                stepper.open(e.Device.SerialNumber);
            }
        }

        void stepper_Detach(object sender, Phidgets.Events.DetachEventArgs e)
        {
            foreach (Stepper stepper in steppers)
            {
                if (stepper.SerialNumber == e.Device.SerialNumber)
                {
                    stepper.close();
                    steppers.Remove(stepper);
                    break;
                }
            }

            stepperStat.Text = steppers.Count + " Steppers Attached";
        }

        void stepper_Attach(object sender, Phidgets.Events.AttachEventArgs e)
        {
            Stepper stepper = (Stepper)e.Device;
            stepper.steppers[0].Engaged = false;
            stepper.steppers[0].CurrentLimit = 1;
            stepper.steppers[0].Acceleration = 1000000;
            stepper.steppers[0].TargetPosition = stepper.steppers[0].PositionMax;
            steppers.Add(stepper);

            stepperStat.Text = steppers.Count+" Steppers Attached";
        }

        Boolean noOff = false;
        private void noOffCheck_CheckedChanged(object sender, EventArgs e)
        {
            noOff = noOffCheck.Checked;
        }

        private void Form1_FormClosing(object sender, FormClosingEventArgs e)
        {
            stopPlayerThreads();
            foreach(Stepper stepper in steppers)
                stepper.close();
        }

        private void channelCmb_SelectedIndexChanged(object sender, EventArgs e)
        {
            channelIndex = int.Parse(channelCmb.SelectedItem.ToString());
        }
    }

    public enum MIDIFormat { Single=0, MultiTrack=1, MultiSong=2 };
    public class MidiFile
    {
        public MidiTrack[] tracks;
        public int speed; //ticks per quarter note
        public int tempo = 1000000; //default tempo 1 second per quarter note (in us)
    }

    public class MidiTrack
    {
        public byte[] trackData;
        public List<String> trackNames = new List<string>();
        public List<NoteEvent> noteData = new List<NoteEvent>();
        public List<int> channels = new List<int>();
        public int tempo = 0;

        public MidiTrack(byte[] trackData, int tempo)
        {
            this.trackData = trackData;
            this.tempo = tempo;
            parseTrackData();
        }

        private void parseTrackData()
        {
            int index = 0;
            MidiEvent lastMidiEvent = 0;
            int lastChannel = 0;

            while(index < trackData.Length)
            {
                int vTime = trackData[index] & 0x7F;
                int shift = 0;
                while ((trackData[index] & 0x80) == 0x80)
                {
                    index++;
                    shift += 7;
                    vTime <<= shift;
                    vTime += trackData[index] & 0x7F;
                }
                index++;

                if (index >= trackData.Length)
                    throw new Exception("Index out of sync!");

                //Find event type = midi, meta, or sysex
                int eventType = trackData[index];

                MidiEvent midiEvent;
                int channel = 0;

                //continuation of last mode - no event type specified.
                if ((eventType & 0x80) == 0x00)
                {
                    midiEvent = lastMidiEvent;
                    channel = lastChannel;
                }
                //new event type
                else
                {
                    index++;
                    if ((eventType & 0xF0) != 0xF0)
                    {
                        channel = eventType & 0x0F;
                        eventType = eventType & 0xF0;
                    }
                    midiEvent = (MidiEvent)Enum.Parse(typeof(MidiEvent), eventType.ToString());
                }

                lastMidiEvent = midiEvent;
                lastChannel = channel;

                switch (midiEvent)
                {
                    //Meta
                    case MidiEvent.MetaEvent:
                        //do this 1st, so our index is valid
                        int MetaTypeIndex = index++;

                        //V-Length
                        int MetaLength = trackData[index] & 0x7F;
                        shift = 0;
                        while ((trackData[index] & 0x80) == 0x80)
                        {
                            index++;
                            shift += 7;
                            MetaLength <<= shift;
                            MetaLength += trackData[index] & 0x7F;
                        }
                        index++;

                        int MetaIndex = index;
                        index += MetaLength;
                        try
                        {
                            MetaType type = (MetaType)Enum.Parse(typeof(MetaType), trackData[MetaTypeIndex].ToString());
                            //got here - it parsed it fine.

                            switch (type)
                            {
                                case MetaType.TrackName:
                                    trackNames.Add(Encoding.UTF8.GetString(trackData, MetaIndex, MetaLength));
                                    break;
                                case MetaType.TempoSetting:
                                    //tempo in microseconds per quarter note
                                    tempo = (trackData[MetaIndex] << 16) + (trackData[MetaIndex] << 8) + trackData[MetaIndex];
                                    //Console.WriteLine("Got a tempo setting: "+tempo);
                                    break;
                                case MetaType.EndOfTrack:
                                    break;
                                case MetaType.TextEvent:
                                case MetaType.CopyrightNotice:
                                case MetaType.InstrumentName:
                                case MetaType.LyricText:
                                case MetaType.MarkerText:
                                    String text = Encoding.UTF8.GetString(trackData, MetaIndex, MetaLength);
                                    Console.WriteLine("Unhandled Meta Event: " + type.ToString() + " \"" + text.Trim()+"\"");
                                    break;
                                case MetaType.TimeSignature:
                                case MetaType.KeySignature:
                                default:
                                    Console.WriteLine("Unhandled Meta Event: " + type.ToString());
                                    break;
                            }
                        }catch{}//just ignore errors here.
                        break;
                    case MidiEvent.NoteOff:
                    case MidiEvent.NoteOn:
                        noteData.Add(new NoteEvent(midiEvent, channel, (vTime * tempo / 1000.0), new byte[] { trackData[index++], trackData[index++] }));
                        if (!channels.Contains(channel))
                            channels.Add(channel);
                        break;
                    case MidiEvent.KeyPressure:
                    case MidiEvent.ControllerChange:
                    case MidiEvent.PitchBend:
                        index += 2;
                        break;
                    case MidiEvent.ProgramChange:
                    case MidiEvent.ChannelPressure:
                        index += 1;
                        break;
                    //Unsupported
                    case MidiEvent.SysEx:
                    case MidiEvent.EndOfSysEx:
                        throw new Exception("Can't handle SysEx events");
                    default:
                        throw new Exception("Can't handle this event type");
                }
            }
        }

        private enum MetaType
        {
            SeqNum = 0x00, TextEvent = 0x01, CopyrightNotice = 0x02, TrackName = 0x03, InstrumentName = 0x04,
            LyricText = 0x05, MarkerText = 0x06, CuePoint = 0x07, MIDIChannelPrefix = 0x20, MIDIPortPrefix = 0x21, 
            EndOfTrack = 0x2F, TempoSetting = 0x51, SMPTEOffset = 0x54, TimeSignature = 0x58, KeySignature = 0x59, 
            SequencerEvent = 0x7F
        };

    }
    public enum MidiEvent
    {
        NoteOff=0x80, NoteOn=0x90, KeyPressure=0xA0, ControllerChange=0xB0, ProgramChange=0xC0, ChannelPressure=0xD0, PitchBend=0xE0,
        SysEx=0xF0, SongPosn=0xF2, SongSelect=0xF3, BusSelect=0xF5, TuneRequest=0xF6, EndOfSysEx=0xF7, TimingTick=0xF8, StartSong=0xFA,
        ContinueSong=0xFB, StopSong=0xFC, ActiveSensing=0xFE, MetaEvent=0xFF
    };

    public class NoteEvent
    {
        public MidiEvent eventType;
        public int channel;
        public double duration;
        public byte[] data;

        public NoteEvent(MidiEvent eventType, int channel, double duration, byte[] data)
        {
            this.eventType = eventType;
            this.channel = channel;
            this.duration = duration;
            this.data = new byte[data.Length];
            for (int i = 0; i < data.Length; i++)
                this.data[i] = data[i];
        }
    }

    public class PlayerObject
    {
        public MidiFile midi;
        public MidiTrack track;
        public int channel;
        public Stepper stepper;
    }
}
